/*** alloc *** (c) J.Kolafa 1991, updated 1995, 2008 ***

 o Sophisticated alloc/free routines 
 o Array range checking
 o Some string stuff (result of sprintf as pointer to a static string)

 WARNING: because of some agressive #defines, alloc.h MUST be 
 #included after all system #include's !!!

 The actual functionality depends on following #define's:
   CHECKHEAP (checking level)
   SDS       (self-defined structures)
   FREENULL  (1: free(NULL) is skipped [new 1/2001, default #ifndef FREENULL]
              0: [approx. old version]: free(NULL) is error)

The following macros are available for CHECKHEAP>=0:

alloc(pointer,size)       allocates  size  of heap memory to  pointer

alloczero(pointer,size)   allocates  size  of heap memory to  pointer
                          and fills the memory by zeros

free(pointer)             deallocates memory pointed to by  pointer
                          and assigns NULL to it
			  #if FREENULL : free(NULL) is skipped
			  #if !FREENULL : free(NULL) is error

In addition, the following macros are available for CHECKHEAP<0:

ralloc(pointer,size)      as  alloc  and  alloczero  and all all alloc's are
ralloczero(pointer,size)  recorded and can be freed at once by  release

release(pointer)          works like in Pascal - all dynamic variables 
                          allocated since  pointer  has been allocated
                          are freed.

  NOTE: Variables allocated by  alloc  must not be freed by  release  and
  those allocated by  ralloc  must not be freed by  free !

The checking level depends on the absolute value of CHECKHEAP :
  0: no check - system  malloc, cmalloc,  and  free  used; DANGEROUS!
  1: heap overflow and bad calls to  free  are checked; safe for use, but 
     no diagnostics is printed
  2: as 1 with extended error messages, possibility of array range check
     and alloc/free debug tracing.
     Use the following global variables:
       AllocRange  Number of bytes allocated before and after any memory 
                   This area is checked for overwrites whenever any 
                   alloc/free routine is called, as well as when
                   checkranges is called
                   The value of AllocRange must be a multiple of 16
                   WARNING: memory consuming and slow
       AllocTrace  If set, all alloc's and free's are traced
     In addition, copyarray contains element size check

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include "int4.h"

/* TYPEOF = typeof in C++ 
   TYPEOF = <nothing> in C
This obfuscation is needed because C smoothly casts void* into <any>* 
but C++ does not and other ways to overcome this lead to many warnings 
in the fundamentalistic gcc/g++ compiler,
*/

#ifdef __cplusplus
#define TYPEOF(X) (typeof(X))
extern "C" {
#else
#define TYPEOF(X) /* no cast needed */
#endif

#ifndef FREENULL
#define FREENULL 1
#endif

#ifndef CHECKHEAP
#define CHECKHEAP -1
#endif

#if CHECKHEAP<-2 || CHECKHEAP>2
#error bad CHECKHEAP
#endif

extern int4 AllocSizeLim;

#include "loop.h"

/***** alloc/free *****/

#if CHECKHEAP == 0 /* ==================================================== 0 */

/***** absolutely no check! *****/

#define alloc(X,Y) X=TYPEOF(X) malloc(Y)
#define alloczero(X,Y) X=TYPEOF(X) calloc(1,Y)

void *myfree(void *f);
#define free(X) X=TYPEOF(X) myfree(X)
#define checkranges(VERBOSE) /* nothing */

#elif CHECKHEAP == 1 || CHECKHEAP == -1 /* ========================== +1,-1 */

/***** heap overflow and bad calls to free are checked *****/

void *mymalloc(int4 size,int zero);
#define alloc(X,Y) X=TYPEOF(X) mymalloc(Y,0)
#define alloczero(X,Y) X=TYPEOF(X) mymalloc(Y,1)

void *myfree(void *f);
#define free(X) X=TYPEOF(X) myfree(X)
#define checkranges(VERBOSE) /* nothing */

#elif CHECKHEAP == 2 || CHECKHEAP == -2 /* ========================= +2,-2 */

/*
  bad calls to alloc and free are diagnosed with extended error messages
  range checks available:
    AllocRange bytes allocated before and after and checked
    AllocRange must be a multiple of 16
    (memory consuming: also a some control structures allocated!)
*/

extern int4 AllocRange, AllocTrace;

void *mymalloc(int4 size,int zero,char *c,char *fn,int line);
#define alloc(X,Y) X=TYPEOF(X) mymalloc(Y,0,#X,__FILE__,__LINE__)
#define alloczero(X,Y) X=TYPEOF(X) mymalloc(Y,1,#X,__FILE__,__LINE__)

void mycheckranges(int verbose,char *fn,int line);
#define checkranges(VERBOSE) mycheckranges(VERBOSE,__FILE__,__LINE__)

void *myfree(void *f,char *c,char *fn,int line);
#define free(X) X=TYPEOF(X) myfree(X,#X,__FILE__,__LINE__)

#endif


/***** ralloc/release *****/

#if CHECKHEAP == -1

void *myralloc(int4 size,int zero);
#define ralloc(X,Y) X=TYPEOF(X) myralloc(Y,0)
#define ralloczero(X,Y) X=TYPEOF(X) myralloc(Y,1)

void *myrelease(void *p);
#define release(X) X=TYPEOF(X) myrelease(X)

#elif CHECKHEAP == -2

void *myralloc(int4 size,int zero,char *c,char *fn,int line);
#define ralloc(X,Y) X=TYPEOF(X) myralloc(Y,0,#X,__FILE__,__LINE__)
#define ralloczero(X,Y) X=TYPEOF(X) myralloc(Y,1,#X,__FILE__,__LINE__)

void *myrelease(void *p,char *c,char *fn,int line);
#define release(X) X=TYPEOF(X) myrelease(X,#X,__FILE__,__LINE__)

#endif


#ifdef SDS
/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Self-defined structures (SDS) contain their size as int at the beginning
% (the size of this int is included into the count)
%
% If A and B are pointers to SDS, then
%   sdsalloc(A,<size in bytes>)  allocates SDS A.  The size is also written to
%                                the first int
%   sdsralloc(A,<size in bytes>) as above with ralloc
%   sdsalloczero,sdsralloczero : as above plus fills the memory by zeroes
%   sdscopy(A,B)  copies *B into *A.  Length (but not overlap...) are checked
%   sdszero(A) fills the contents of *A by zeroes (excl. the size)
%
% example:
%   typedef struct { int size; double a[10]; } xtype;
%   xtype X={sizeof(X)}, *A;
%   sdszero(&X);
%   sdsalloc(A,sizeof(xtype));
%   sdscopy(A,&X);
%
% to be used as include file
% required: 
%   macro Error(char *msg) and include module "alloc.c"
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#define sdscopy(X,Y) { \
  if (*(int*)(X) < *(int*)(Y)) ERROR(("SDS:lengths")) \
  else copy((int*)(X)+1,(int*)(Y)+1,*(int*)(Y)-sizeof(int)); }
/*.....  else if (abs((char*)(X)-(char*)(Y)) < *(int*)(X)) Error(sdserrovl); \*/

#define sdszero(X)  { \
  if (*(int*)(X) < sizeof(int)) ERROR(("SDS:bad length")) \
  else memset((int*)(X)+1,0,*(int*)(X)-sizeof(int)); }

#define sdssize(X) (*(int*)(X))

#define sdsalloc(X,Y) { alloc((X),Y); *(int*)(X)=Y; }
#define sdsalloczero(X,Y) { alloczero((X),Y); *(int*)(X)=Y; }

#define sdsralloc(X,Y) { ralloc((X),Y); *(int*)(X)=Y; }
#define sdsralloczero(X,Y) { ralloczero((X),Y); *(int*)(X)=Y; }

#endif /* SDS */


/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  string support
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
*/

void stringinit(int len,int n);
/*
  len=max length of strings
  n=max number of strings available
  (if stringinit is not called, stringinit(80,3) is assumed) 
*/

char *string(const char *format,...);
/* 
  returns pointer to a static string containing the result of
  sprintf(<static string>,format,data)
  Example:
    FILE *f=fopen(string("config.%d",number));
*/

#ifdef __cplusplus
}
#endif

/* added 7/2000 */
#define allocarray(X,Y) alloc((X),(Y)*sizeof((X)[0]))
#define rallocarray(X,Y) ralloc((X),(Y)*sizeof((X)[0]))
#define allocarrayzero(X,Y) alloczero((X),(Y)*sizeof((X)[0]))
#define rallocarrayzero(X,Y) ralloczero((X),(Y)*sizeof((X)[0]))

#define allocone(X) alloc((X),sizeof((X)[0]))
#define rallocone(X) ralloc((X),sizeof((X)[0]))
#define alloconezero(X) alloczero((X),sizeof((X)[0]))
#define ralloconezero(X) ralloczero((X),sizeof((X)[0]))

/* added 1/2001 */
#if CHECKHEAP == 2 || CHECKHEAP == -2
#define copyarray(X,Y,Z) do { \
  if (sizeof((Y)[0]) != sizeof((X)[0])) ERROR(("copyarray: element sizes disagree")) \
  copy((X),(Y),(Z)*sizeof(Y[0])); } while (0)
#else
#define copyarray(X,Y,Z) copy((X),(Y),(Z)*sizeof(Y[0]))
#endif

/* added 12/2003 */
#define arrayzero(X,Y) memset((X),0,(Y)*sizeof((X)[0]))
/* WARNING: this is not ANSI C portable, although I do not know about
   architecture where this trick would fail */
