/***************** STATICS.C ******* (c) J. Kolafa ****************/
/*                                                                */
/*      autocorrelation analysis of serially correlated data      */
/*       STATICS.PLI and STATICS.PAS 1990,  STATISTI.C 1991       */
/*                                                                */
/*  update 02/07 bug (stuck if very large number) fixed           */
/*  update 08/06 function StaStdErr added                         */
/*  update 08/05 to produce files of time covariances             */
/*  update 10/96 to produce files of time correlation functions   */
/*  update 11/95 for better portability, old version=statics0.c   */
/*  update 02/92 2^n subaverage (block) analysis added            */
/*                                                                */
/******************************************************************/

/*
  StaMean, StaN, StaVar accept '\t' at the end of the name as a wildcard * 
*/

#include "ground.h"
#include "alloc.h"
#include "varfile.h"
#include "statics.h"

#define StaNameLen 24 /* should be multiple of 8 */

/*** StaType and StaPtr ***/
typedef struct { double FirstVal,LastVal,SumCov; } StaType;
typedef struct QStruct *StaPtr;
struct QStruct {
  StaPtr    next;     /* -> to the next set of measurements */
  double    align_me; /* double-compatible alignment on different              
                         architectures */
  int4      no;       /* # of measurements */
  int4      lag;      /* for original data */
  int4      blklag;   /* for block-averaged data */
  int4      noblk;    /* # of blocks */
  char      name[StaNameLen];
  StaType   s[1/*var len*/]; };

/***
  format of data in StaPtr->s[] :

  subscript_______________________contents_________________
  0                               min x, max x, SUM x^3
  1 .. 1+lag                      statistics upto lag of original data
  2+lag .. 2+lag+blklag           statistics of averages by 2 pieces of data
  3+lag+blklag .. 3+lag+2*blklag  statistics of averages by 4 pieces of data
  etc.                            etc.

  one block of statistics (of length of lag+1 or blklag+1) :

  rel.index___0_____________1______________2_________..._____lag
  FirstVal  SUM y          y[1]           y[2]
  LastVal     0            y[n]           y[n-1]
  SumCov    SUM y^2   SUM y[i]*y[i-1]  SUM y[i]*y[i-2]

  where y is x or a subaverage of 2,4,... x's  and n the number of y's recorded
***/

/*** StaInfo ***/
static struct {
  unsigned lag;    /* lag for original data */
  unsigned blklag; /* lag for data averaged in blocks of lengths of */
  unsigned noblk;                                 /* 2,4,...2^noblk */
  StaPtr   first;  /* first item in the list */
  StaPtr   last;   /* last item in the list (see StaFind) */
  char tname[StaNameLen];
                   /* truncated name (see StaFind) */
  } StaInfo = { 32, 2, 12, NULL, NULL, "<ndef>" };


/*** local functions ***/

static unsigned StaDblSize(int lag,int noblk,int blklag) /* # of doubles */
{
  return (lag+2+noblk*(blklag+1))*3;
}

static unsigned StaSize(unsigned qdsize) /* size in bytes */
{
  return sizeof(struct QStruct) + sizeof(double)*(qdsize-3);
}

int StaError; /* set to 1 if name not found */

static StaPtr StaFind(const char *name,int msglev)
{
  StaPtr Q=StaInfo.first;
  char *tab=strchr(name,'\t');
  int l=tab?tab-name:0;

  StaError=0;

  /* truncate a too long name */
  if (strlen(name)>=StaNameLen) {
    copy(StaInfo.tname,name,StaNameLen);
    StaInfo.tname[StaNameLen-1]=0; }
  else
    strcpy(StaInfo.tname,name);

  while (Q) {
    if ( (l && memcmp(Q->name,StaInfo.tname,l)==0)
      || (l==0 && strcmp(Q->name,StaInfo.tname)==0) ) break; /* name found */
    StaInfo.last=Q; Q=Q->next; }

  if (Q==NULL) {
    StaError++;
    if (msglev) prt("WARNING StaFind: %s not found",StaInfo.tname); }

  return Q;
}

static double StaAddBlk(StaType *q, int4 n, int lag)
{
  int i;
  double x=q->LastVal; /* input value = block subaverage */

  q->FirstVal += x; /* sum x */
  if (n<=lag) q[n].FirstVal=x;
  loopto (i,0,lag) q[i].SumCov += x*q[i].LastVal;
  for (i=lag; i>0; i--) q[i].LastVal = q[i-1].LastVal;
  q->LastVal=0;

  return x;
}

static FILE *gnu;
static int precision=0;
static char fmt[]=" %3u%7.4f %6.3f";

static int lstd;
static double *laststd,av;
static int cov=0;

static double StaPrtBlk(int pr,StaType *q, int4 n, unsigned lag, unsigned blk)
{
  double sf=q->FirstVal;
  double sl=q->FirstVal;
  double t,cor=1,vari,varn,mx=0;
  unsigned i,m=n/3;

  if (n<2) return 0;
  blk /= n;
  if (m>lag) m=lag;
  av=q->FirstVal/n;
  vari=q->SumCov/n - av*av;
  if (vari<0) vari=0.0;
  varn=vari/(n-1)/Sqr(blk);
  if (vari==0) return 0;
  if (cov) cor=vari;
  t=cor;

  for (i=0;;) {

    if (gnu && pr) {
      if (cov)
        fprintf(gnu,"%3u %.8g\n",i,cor);
      else
        fprintf(gnu,"%3u %9.6f\n",i,cor); }
    else {
      double e;
  
      if (pr) {
	fmt[0]=' ';
	if (i%3==0) {
	  _n
	  if (blk!=1) fmt[0]='*'; }
	fmt[13]='2'+(t<100)+precision*2;
	fmt[5]='7'+precision*2;
	prt_(fmt,i,cor,t); }

      if (t<0) e=0; else e=sqrt(varn*t);
      Max(mx,e)

      if (laststd && i==1) laststd[lstd++]=e;

      if (pr) {
	if (precision)
	  prt_("%13.5e",e);
	else {
	  /* e-format is not compatible => it is created here from f and i */
	  int p=0;

	  if (e) {
	    while (e<0.995) { e*=10; p--; if (p<-999) break; }
	    while (e>9.95) { e/=10; p++; if (p>999) break; } }
	  if (abs(p)>999) prt_(" ???????");
	  else prt_("%4.1fe%c%i%i",e,p>=0?'+':'-',abs(p)/10,abs(p)%10); } } }
      
    if (++i>m) break;
  
    sf=sf-q[i].FirstVal;
    sl=sl-q[i].LastVal;
    cor=(q[i].SumCov-sf*sl/(n-i))/(n-i);
    if (!cov) cor/=vari;
    t+=cor+cor; }

  return mx;
}


/*** user functions ***/

void StaSet(int lag,int blklag,int noblk) /************************** StaSet */
{
  StaInfo.lag=lag; StaInfo.blklag=blklag; StaInfo.noblk=noblk;
}

void StaAdd(const char *name, double x) /**************************** StaAdd */
/*** next value x of quantity tx is processed ***/
{
  StaPtr Q=StaFind(name,0);
  unsigned i;
  int4 shift;
  StaType *q;

  if (Q==NULL) {
    /*** free area for a new data set appended to the list ***/
    int size=StaDblSize(StaInfo.lag,StaInfo.noblk,StaInfo.blklag);

    alloc(Q,StaSize(size));
    Q->next=NULL;
    Q->no=0;
    Q->lag=StaInfo.lag;
    Q->blklag=StaInfo.blklag;
    Q->noblk=StaInfo.noblk;
    copy(Q->name,StaInfo.tname,StaNameLen);
    if (StaInfo.first==NULL) StaInfo.first=Q;
    else StaInfo.last->next=Q;
    loop (i,2,size) ((double*)Q->s)[i]=0.0;
    Q->s[0].FirstVal=Q->s[0].LastVal=x; }

  /*** x added to the data set ***/

  Q->no++;

  q=Q->s;
  if (x<q->FirstVal) q->FirstVal=x; /* MIN(x) */
  if (x>q->LastVal) q->LastVal=x;   /* MAX(x) */
  q->SumCov += x*x*x;               /* SUM(x^3) */

  q++;
  /*** statistics of original data (no subaverages) ***/

  q->LastVal=x;
  StaAddBlk(q,shift=Q->no,Q->lag);

  q += Q->lag+1;
  /*** statistics of data in 2,4,8,... blocks ***/

  loopto (i,1,Q->noblk) {
    x = q->LastVal += x;
    if (shift & 1) break;
    shift >>= 1;
    StaAddBlk(q,shift,Q->blklag);
    q += Q->blklag+1; }

} /* StaAdd */


void StaFree(void) /*********************************************** StaFree */
/* clears all measurements (not necessary in the first call) */
{
  StaPtr Q;
  while (StaInfo.first) {
    Q=StaInfo.first;
    StaInfo.first=StaInfo.first->next;
    free(Q); }
} /* StaFree */


double StaPrint(const char *name,const char *how) /**************** StaPrint */
/*
  Prints statistics listing of one quantity.  String how determins the mode:
  NULL   compact table with c_t, tau, StDev for error analysis printed to out
  "+"    as above with more decimal digits
  "-"    t:c_t table printed to stdout
  other: t:c_t table printed to file how.name.tcf (if name contains
             non-alpha characters, it is truncated)
  NEW: * in front of HOW will cause covariances to be printed
  WARNING: `how' logic is awfuls and should be changed...!!!
*/
{
  StaPtr Q=StaFind(name,1);
  double av,vari /*,varn,sdev0*/ ;
  int from=0,to;
  double mx=0;
  unsigned i;
  int4 shift,prec0;
  StaType *q;

  if (!Q) return 0;

  if (how) if (how[0]=='*') cov=1,how++;

  prec0=precision;
  q=Q->s;

  if (how)
    if (strcmp(how,"+")) {
      char *fn,*c1;
      const char *c2;

      if (strcmp(how,"-")) {
        alloc(fn,strlen(how)+strlen(name)+6);
        strcpy(fn,how);
#ifndef DOS
        if (*fn) strcat(fn,".");
#endif
        for (c2=name; *c2==' '; c2++);

        /* should truncate to 8 chars for DOS, but I ignore it */
        for (c1=fn+strlen(fn); *c2; c2++)
          /* ignore bad characters (changed 8/05) */
          if (isalnum(*c2) || strchr("_-~!@#.",*c2)) *c1++=*c2;
        strcpy(c1,cov?".cov":".tcf");
        if ( !(gnu=fopen(fn,"wt")) ) ERROR(("open %s",fn))
        free(fn); }
      else
        gnu=stdout,precision=1; }
    else gnu=NULL,precision=1;
  else
    gnu=NULL,precision=0;

  if (cov) precision=1;

  if (gnu)
    fprintf(gnu,"# %-20s  No=%lu  range=[%.*g,%.*g]=%.*g\n",
           Q->name,
            (long unsigned)Q->no,
            7+precision*8,q->FirstVal,
            7+precision*8,q->LastVal,
            7+precision*8,q->LastVal-q->FirstVal);
  else
    prt("%-20s  No=%lu  range=[%.7g,%.7g]=%.7g",
        Q->name,
        (long unsigned)Q->no,
        q->FirstVal, q->LastVal, q->LastVal-q->FirstVal);

  /*** statistics of original data (no subaverages) ***/
  q++;

  av=q->FirstVal/Q->no;
  if (gnu) fprintf(gnu,"#  Mean = %.9g", av);
  else prt_("  Mean = %.*g", 9+6*precision,av);
  if (Q->no<2) {
    if (!gnu) _n goto ret; }
  vari=q->SumCov/Q->no - av*av;
  if (vari<0) vari=0.0;
  /* varn=vari/(Q->no-1);
     sdev0=sqrt(varn);
     prt("    Var = %.9g    StDev0 =%11.4e",vari,sdev0); */
  if (gnu) fprintf(gnu,"    Var = %.9g\n",vari);
  else prt("    Var = %.*g",9+6*precision,vari);
  if (vari==0 || Q->no<3) {
    if (!gnu) _n goto ret; }

  if (gnu)
    if (cov) 
      fprintf(gnu,"#\n#_t__Cov[t]\n");
    else
      fprintf(gnu,"#\n#_t____c[t]\n");
  else
    loopto (i,1,3)
      if (precision) prt_("  _t____c[t]___1+2tau_______StDev_");
      else prt_("  _t__c[t]__1+2tau__StDev_");

  mx=StaPrtBlk(1,q,shift=Q->no,Q->lag,Q->no);

  if (gnu) goto ret;

  q += Q->lag+1;
  /*** statistics of data in 2,4,8,... blocks ***/

  if (Q->noblk) allocarrayzero(laststd,Q->noblk);
  lstd=0;

  loopto (i,1,Q->noblk) {
    shift >>= 1;
    StaPrtBlk(1,q,shift,Q->blklag,Q->no);
    q += Q->blklag+1; }

  if (Q->noblk>2) {
    if (laststd[from]==0) from++;
    if (laststd[from]==0) from++;
    if (laststd[from]) {
      to=from;
      while (laststd[to] && to<Q->noblk) to++;
      loop (i,from,to) Max(mx,laststd[i]) } }
     
  if (Q->noblk) free(laststd);
  laststd=NULL;

  _n
  if (mx) {
    static int pass=1;
    if (pass) {
      pass=0;
      prt("#____average____stderr__rel.stderr__no__=name="); }
    prt(precision ? "# %.15g %.6g %.6g %d =%s="
        : "# %.12g %.3g %.3g %d =%s=",
        av,mx,mx/av,Q->no,name); }
  _n

 ret:
  cov=0;
  if (gnu && gnu!=stdout) fclose(gnu);

  precision=prec0;

  return mx;
} /* StaPrint */


void StaPrintAll(const char *how) /***************************** StaPrintAll */
/* see StaPrint for `how' */
{
  StaPtr Q=StaInfo.first;
  while (Q) {
    StaPrint(Q->name,how);
    Q=Q->next; }
} /* StaPrintAll */


static void StaXSave(char *fn)
/* saves all measurements to file named fn */
{
  StaPtr Q=StaInfo.first;
  VarOpen(fn,"wb");
  while (Q) {
    VarPut(&Q->no,
           StaSize( StaDblSize(Q->lag,Q->noblk,Q->blklag))
           -(int)((char*)&Q->no-(char*)Q) );
    Q=Q->next; }
} /* StaXSave */


void StaSave(char *fn) /******************************************** StaSave */
{
  StaXSave(fn);
  VarClose();
}


void StaKeySave(char *fn,int4 key) /******************************** StaSave */
{
  StaXSave(fn);
  KeyClose(key);
}

void StaLoad(char *fn) /******************************************** StaLoad */
/*
  loads all measurements from file named fn until EOF 
*/
{
  StaPtr Q=StaInfo.first;
  unsigned oldsize,size;

  VarOpen(fn,"rb");
  while (!VarEOF) {

    /* find list end */
    Q=StaInfo.first;
    while (Q) {
      StaInfo.last=Q;
      Q=Q->next; }

    /* normal read - no conversion */
    alloc(Q,oldsize=VarSize+(int)((char*)&Q->no-(char*)Q));
    VarRead(&Q->no,VarSize); 
    size=StaSize(StaDblSize(Q->lag,Q->noblk,Q->blklag));
    if (size != oldsize) {
      ERROR(("StaLoad: size read=%u != expected = %u (try STAOLDFMT?)",
             oldsize,size)) 
      return; }

    /* incl. to list end */
    if (StaInfo.first==NULL) StaInfo.first=Q; 
    else StaInfo.last->next=Q;
    Q->next=NULL; }

  VarClose();
} /* StaLoad */


int4 StaN(const char *name) /****************************************** StaN */
/* # of recorded measurements of quantity named name */
{
  StaPtr Q=StaFind(name,1);
  if (Q) return Q->no; else return 0;
} /* StaN */


double StaMean(const char *name) /********************************** StaMean */
/* mean value of quantity named name */
{
  StaPtr Q=StaFind(name,1);

  if (Q) return Q->s[1].FirstVal/Q->no; else return 0.0;
} /* StaMean */


double StaVar(const char *name) /************************************ StaVar */
/* variance of quantity named name */
{
  StaPtr Q=StaFind(name,1);

  if (Q==NULL) return 0.0;
  else {
    double av=Q->s[1].FirstVal/Q->no;
    return Q->s[1].SumCov/Q->no - av*av; }
} /* StaVar */


double StaStdErr(const char *name) /****************************** StaStdErr */
/*
  estimated standard error: the performance depends on StaSet() settings!
  (the same as StaPrint last line)
*/
{
  StaPtr Q=StaFind(name,1);
  double av,vari;
  int from=0,to;
  double mx;
  unsigned i;
  int4 shift;
  StaType *q;

  if (!Q) return 0;

  q=Q->s;

  /*** statistics of original data (no subaverages) ***/
  q++;

  av=q->FirstVal/Q->no;
  vari=q->SumCov/Q->no - av*av;
  if (vari<0) vari=0.0;
  if (vari==0 || Q->no<3) return 0;

  mx=StaPrtBlk(0,q,shift=Q->no,Q->lag,Q->no);

  q += Q->lag+1;
  /*** statistics of data in 2,4,8,... blocks ***/

  if (Q->noblk) allocarrayzero(laststd,Q->noblk);
  lstd=0;

  loopto (i,1,Q->noblk) {
    shift >>= 1;
    StaPrtBlk(0,q,shift,Q->blklag,Q->no);
    q += Q->blklag+1; }

  if (Q->noblk>2) {
    if (laststd[from]==0) from++;
    if (laststd[from]==0) from++;
    if (laststd[from]) {
      to=from;
      while (laststd[to] && to<Q->noblk) to++;
      loop (i,from,to) Max(mx,laststd[i]) } }

  if (Q->noblk) free(laststd);
  laststd=NULL;
  cov=0;

  return mx;
} /* StaStdErr */
