/* /usr/bin/make hsmd

********** MD of pure HS ********** (c) J. Kolafa 2002-2008 **********

This program is free software (c) Jiri Kolafa and may be distributed under
the terms of the GNU General Public Licence.  See file COPYING for details.

**********************************************************************

This program combines the ideas of the linked-cell list method both in
space and time.  With the exception of the `hash table', these ideas
are similar to those published by:
[1] J.J. Erpenbeck and W.W. Wood: MD Techniques for Hard-Core Systems.
    in Statistical Mechanics, part B, ed. B.J. Berne (Plenum Press, 1977)
[2] J.J. Erpenbeck and W.W. Wood: MD Calculation of the Hard-Sphere
    Equation of State.  J. Stat. Phys., 35, 321(1984)

Basic object is `struct HS_s' representing a hard sphere.  There is
array `cfg' of HS.

In addition, there is a 3D array `box' of cells containing linked
lists of HS_s.  When collisions or RDF are calculated, cells around
one chosen cell are drawn according to the precalculated list of cells
(nbr, rdfnbr); there are NSUBCELL^3 such lists based on the position
of central HS in its cell.

Time control is accomplished via a (partially) time-ordered LIST OF
EVENTS; typical examples of events are:
  - collision of a pair of spheres
  - sphere leaves the cutoff area
  - measure (RDF, ...)
  - stop simulation
This list of events is organized as an ARRAY OF TIME SLICES `tlist'
plus extra slice `longt' for even longer times.  For instance, in time
slice [0,dt) there are events (implemented as a linked list) that are
possibly going to happen at times t, 0<=t<dt, in time slice [dt,2*dt)
there are events to happen at times dt<=t<2*dt, etc.  We use such time
units so that dt=1.

One simulation step begins with the selection of the first event that
is going to happen: In the first time slice, the list is searched and
the minimum time is found.  This event is performed.  If this event is
`collision of spheres', it is necessary to recalculate the collision
times of the affected spheres.  This is done using the following
algorithm for both spheres:
  1/ for all spheres within cutoff new collision times are calculated
  2/ old collisions are removed from the list of events
  3/ new collisions are scheduled
  4/ time when the sphere moves out of one half of the cutoff distance
     is scheduled as event `recalculate collisions'

To perform step 2/, it is necessary to have a mechanism to remove the
scheduled collisions, i.e., to find (efficiently) the events in the
list.  For small systems, this can be easily implemented by a matrix
(2D array) indexed by the spheres.  For large systems this would take
too much memory, but note that the matrix is sparse.  We therefore use
a HASH TABLE METHOD instead.  The hash table is a 1D array of lists of
pointers to events.  Each sphere has a hash value assigned, the index
in the hash table is the sum of the hash values of both spheres
(modulo hash table size), and a list starting at the selected element
is searched for the pair.  Pointer to event `recalculate collision' is
stored directly at the sphere.

During the simulation, the events from the first time slice are
performed.  As soon as the first slice becomes empty, all slices are
shifted so that the second slice becomes the first, etc.  In addition,
the appropriate events from `longt' are moved to the last slice.  This
is implemented by advancing indices along with to the pointer to the
table.  Ond only once a while, when this `sliding window' (of length
Nlist) moves out of the allocated area (of length Nlist+Nlistshift), a
physical move is done.

Cutoff optimization:
^^^^^^^^^^^^^^^^^^^^
Too small cutoff causes too often recalculations of collisions and the
ratio updates/collisions (column 5=upd in SIMNAME.cp) is large.  Too
large cutoff means that many calculated collisions are wasted (the
affected spheres collide with other spheres).  The optimum
updates/collisions ratio is roughly 0.5--1.  For dense systems the
optimum cutoff is about 1.2 and it increases with decreasing density.

Time slice optimization:
^^^^^^^^^^^^^^^^^^^^^^^^
The time unit used by the program depends on the kinetic energy K=E/2.
The user may set variable E and see how many events are in one time
slice of width 1.  It is more convenient to set variable slice to the
desired number of events and E will be approximately calculated from
EOS.  In addition, the time table length Nlist should be enough so
that there are not too many events outside the array (in list
`longt'), but not too extensive to avoid cache swapping.

Caveats and bugs:
^^^^^^^^^^^^^^^^^
All time data in structure dt are integers for standard HS version (cf.
dttype).  Then the maximum time is 2^31.  This may cause problems mainly 
for long simulations with not too many particles.
Workaround: use larger slice or use dt.stop=0 (and interrupt simulation
by ctrl-c or kill 2).

For the sake of optimization, weird units are used internally.
The time unit = time slice width, length unit = cell size.
And finally mass unit=1.
The conversion is:
  sigma=lengthunit=ncell/L
  timeunit=sigma*sqrt(E/N)  (E=2*Ekin)
  energyunit=N/E  

The RDF cannot be measured to r=L/2 but to a distance by a cell and subcell
less.  It can be measured to L-sigma by a LONG version, but inefficiently 
(no linked-cell list method).
*/

#ifndef NSUBCELL
#define NSUBCELL 6
#endif
/*
  subdivision of cells for cell neighbors' tables
  the larger, the finer division and faster run (unless memory
  required runs out of resources incl. cache)

  WARNING: memory required is approximately:
    4/3*PI*(rdf.cutoff^3 + cutoff^3)*NSUBCELL^3*sigma^3*12B
  for large rdf.cutoff:
    (rdf.cutoff*NSUBCELL)^3/7000 MB
*/

/*
#define LONG
SPECIAL for RDF measured upto L:
- good for small N
- VERY INEFFICIENT for large N (if rdf.cutoff<<L/2)
*/

/*
#define SWPERT 4
- special for monitoring square-well perturbation terms,
  (but does not hurt if turned on)
- the value of SWPERT is the maximum power monitored
- useful range: SWPERT=3,4,5,6, recommended 4
WARNING: format of rdf files is changed and not compatible with
  non-SWPERT version
*/

/*
#define DEBUG 1e-12
  DEBUG MODE: error of tests, #include "hsmdtest.c"; runs much slower
  (some critical tests are active even if #undef DEBUG) */

/*
#define NOCLOCK
  using time() in s instead of clock() */

#define EPS 3e-13 /* - error of other than DEBUG tests;
                      - small time to schedule consecutive events
                      recommended: about 1000*mashine presision
                      (IEEE double: mashine presision=2.220446e-16 */

#include "ground.h"
#include "alloc.h"
#include "rndgen.h"
#include "statics.h"
#include <time.h>
#include <unistd.h>

typedef double real;

#ifdef SWPERT
typedef real dttype;
#if SWPERT<3
#define SWPERT 3
#endif
#else
typedef int dttype;
#endif

#include "vector3d.h"

typedef int intvector[3];

int Nlist=0; /* size of tlist */
int Nlistshift=0; /* extra length of tlist, to optimize reading frame shift */
real t; /* the global time measured from toffset */
real toffset; /* global time offset */
real sigma,sigmaq; /* sigma=HS diameter, sigmaq=sigma^2 */
int ncell; /* cells/box */
real Tprotocol=0.1; /* with init=3,4: should be 0<Tprotocol<1.4
                        large value: faster T decrase while MC annealing
                        smaller value: slower decrase while MC annealing
                        Tprotocol=0.1 is reasonable compromise good
                        for large densities */

struct HS_s {
  struct HS_s *next,
              **cell; /* head list - location in box[][][] */
  vector r,v;
  intvector ir; /* cell coordinates */
  real t; /* time at which r,v are valid (from toffset); it may differ
              from the global time t - see advance() and syncHS() */
  int hash; /* hash value, to be used for looking in pair[] */
  struct event_s *update; /* pointer to the update event for this sphere */
  }
  ****box, /* box[ncell][ncell][ncell] = pointer to HS list */
  *cfg, /* [N], in normal order, is array of HS  */
  *UPDATE,*MEASURE,*PLB,*CFG,*RESET,*BLOCK,*SAVE,*STOP;
                  /* denotes other events than collisions */

#define NUPDATE 8

char *eventname[NUPDATE]={
  "UPDATE", /* this must be the 1st one */
  "MEASURE","PLB","CFG","RESET","BLOCK","SAVE", /* these saved */
  "STOP" /* this must be the last one */ };

int N=1000; /* number of spheres */
real L,Lh; /* box size, L/2 in prog. units (i.e., cell size=1) */
real userL; /* L in user units (i.e., sphere sigma=1) */
real cutoff=1.2; /* in sigma - to be transformed to prog. units */
real cq; /* cutoff^2 */
real collisionrange; /* in which collisions are searched */
real rho=0.92;
int verbose=0;

struct collisions_s {
  int block,sumblock,update;
  real virial; } collisions;

int NOTIME=0x7fffffff;

#ifdef DEBUG
void ovltest(char *title);
void prtsphere(struct HS_s *A);
void checklist(char *title,int p);
#endif

#define Ntot N /* compatibility with mix.c */
#include "hsmdlist.c"

#ifdef DEBUG
#include "hsmdtest.c"
#endif

#include "hsmdcoll.c"

void updatecollisionlist(real tt,                  /*** updatecollisionlist */
                          struct HS_s *A,struct HS_s *B)
/*
  both the event list tlist[] and its reference in the hashed pair table
  pair[] are updated to the new time t+tt for collision A vs. B
  if tt==NOTIME, the event is only removed
*/
{
  int it,ihash;
  struct event_s **head,*l;
  struct pair_s *c,**cprevptr;
  struct HS_s *aux;

  /* normalize order (will simplify search later) */
  if (A>B) aux=A,A=B,B=aux;

  if (tt==NOTIME)
    head=NULL; /* event will be removed */
  else {
    tt+=t;
    if (tt>=(double)0x7fffffff) /* MAXINT ?! */
      head=&longt;
    else {
      it=(int)tt;
      if (it<it1) head=&tlist[it];
      else head=&longt; } }

#ifdef DEBUG
  if (head && it<it0) ERROR(("updatecollisionlist: it=%d it0=%d t=%g tt=%g",it,it0,t,tt))
#endif

  /* sparse 2D array of sphere-sphere pairs via hash table */
  ihash=(A->hash+B->hash)%nhash;
  for (cprevptr=&pair[ihash], c=*cprevptr; c; c=*cprevptr) {
    if (A==c->A && B==c->B) break;
    cprevptr=&c->next; }

  if (c) {

    l=c->l;

    removeevent(l);

    if (!head) {
      /* erasing the event */
      free(l);
      /* removing its pointer from the hash table + erasing it*/
      *cprevptr=c->next;
      free(c);
      return ; } }

  else {
    if (!head) return; /* nothing to do - no old event, no new event */
    /* new tlist and hashed pair table items */
    allocone(l);
    l->A=A; l->B=B;
    allocone(c);
    c->A=A; c->B=B;
    c->l=l;
    c->next=pair[ihash];
    pair[ihash]=c; }

  c->t=l->t=tt;

  /* including event l into tlist */
  l->next=*head; // ERRORs detected here: l=NULL; head cannot be accessed
  l->prev=NULL;
  if (*head) (*head)->prev=l;
  *head=l;
}

#include "rndgauss.c"
#include "hsmdsys.c"
#include "hsmdnbr.c"

void collide(struct HS_s *A,struct HS_s *ignore) /**************** collide */
/*
  update all collisions affected by sphere *A
  sphere upto cutoff are checked around sphere A but sphere ignore
*/
{
  real tt;
  struct HS_s ***Bx,**By,*B,*Bnext;
  struct nbr_s *nbr;
  int x,y,z;

#ifdef DEBUG
  if (!box[A->ir[0]][A->ir[1]][A->ir[2]])
    ERROR(("box[%d][%d][%d] empty - should have a sphere",
           A->ir[0],A->ir[1],A->ir[2]))
  if (A->t != t) ERROR(("A->t=%g != t=%g",A->t,t))
#endif

  /* note: exact rounding A->r into A->ir needed, cf. advance() */

  nbr=&nbrs[(int)((A->r[0]-A->ir[0])*RSUBCELL)]
           [(int)((A->r[1]-A->ir[1])*RSUBCELL)]
           [(int)((A->r[2]-A->ir[2])*RSUBCELL)];

  loop (x,nbr->fromx,nbr->tox) {
    Bx=box[(A->ir[0]+x)%ncell];
    loop (y,nbr->x[x].fromy,nbr->x[x].toy) {
      By=Bx[(A->ir[1]+y)%ncell];
      loop (z,nbr->x[x].y[y].fromz,nbr->x[x].y[y].toz)
        for (B=By[(A->ir[2]+z)%ncell]; B; B=Bnext) {
          Bnext=B->next;
          if (B!=ignore && B!=A) {
            tt=collisiontime(A,B);
            updatecollisionlist(tt,A,B); } } } }

  tt=collisionrange/sqrt(SQR(A->v));

  /* update the `UPDATE' event for sphere A */
#ifdef DEBUG
  if (!A->update) ERROR(("A->update should have been allocated in advance"))
  if (A->update->A!=UPDATE || A->update->B!=A)
    ERROR(("unexpected A->update: A=%x [%d] B=%x [%d] UPDATE=%x",
           A->update->A,A->update->A-cfg,A->update->B,A->update->B-cfg,UPDATE))
#endif

  rescheduleevent(tt,A->update);
}

#include "hsmdinit.c"
#include "hsmdmeas.c"

#include "hsmdfile.c"

real toffset;

void resettime(void) /******************************************* resettime */
/*
  shifts all times by dt=(int)t back and recalculates all time info
  this should be done once a while to avoid rounding errors in time
  differences
*/
{
  int idt=t;
  real dt=idt;
  struct event_s *e;
  struct pair_s *p;
  int i;

  toffset+=dt;
  lastsynctime-=dt;

  t-=dt;
  loop (i,0,N) cfg[i].t-=dt;
  loop (i,it0,it1) looplist (e,tlist[i]) e->t-=dt;
  looplist (e,longt) e->t-=dt;
  loop (i,0,nhash) looplist (p,pair[i]) p->t-=dt;

  it0-=idt; it1-=idt; tlist+=idt;

  prt("time reset: toffset=%g dt=%g new t=%g",toffset,dt,t);
}

struct event_s *controlevent(real tt,struct HS_s *A) /******** controlevent */
/*
  if tt!=0 then new control event is created and scheduled at t+tt
  not good for collision event
*/
{
  struct event_s *e=NULL;

  if (tt) {
    allocone(e);
    e->A=A;
    scheduleevent(tt,e); }

  return e;
}

real gausscorr(int N) /****************************************** gausscorr */
{
#if 1
  return exp(lgamma((DIM*(N-1)+1)/2.)-lgamma(DIM*(N-1)/2.))/sqrt(DIM*N/2.);
#else
  /* asymptotic expansion with error ~ 0.0551/N^4 */
  return 1-(7./12+(47./288+301./3456/N)/N)/N;
#endif
}

int main(int narg,char **arg) /*************************************** main */
{
  char *fn="<stdout>";
  int init=0,i,j;
  int LN[DIM]={0,0,0}; /* for init=5..8: # of elementary cells per box
              0=min (max filled lattice) and random vacancies (if any) */
  double rnucleus=0;
  double lastcpu=0,sumcpubutfirst=0;
  real eta;
  real x=0,rhos=1.041,rhof=0.943,V=0; /* aux. quantity */
  int n=0;
  real E=0,Z;
  real slice=20;
  int lag=20;
  struct {
    dttype measure; /* measure Z and rdf time interval */
    dttype plb;     /* write playback frame */
    dttype cfg;     /* write configuration */
    dttype reset;   /* time offset reassignment (not to cumulate rounding err) */
    dttype block;   /* writing rdf block */
    dttype save;    /* save all (restartable checkpoint) */
    dttype stop;    /* stop run */
    } dt = {100,0,0,100000,100000,100000,0};
  int seed=0;
  clock_t clock0=0,clockt=0;
  real timeunit;
  double Zmin=0,Zmax=0;

  initscroll(0);

  if (narg<2) {
    /* info on NSUBCELL ... LONG also printed to .prt */
    fprintf(stderr,"*** MD of pure HS *** (c) J. Kolafa 2002 *** licence: GPL ***\n\
Compile-time options:\n  NSUBCELL=%d EPS=%g DEBUG=%g"
#ifdef LONG
" LONG (rdf for r>box/2)"
#else
" !LONG (rdf for r<box/2-cell)"
#endif
"\n\
Call by:\n\
  hsmd SIMNAME [init [verbose]]\n\
input file = SIMNAME.get, output files = SIMNAME.{prt,all,plb,cfg,g,res,sta}\n\
init: overrides variable `init' in SIMNAME.get:\n\
  init=0: continue from saved point\n\
  init=1: read cfg, new measurements, append plb+cp\n\
  init=2: read cfg, new measurements\n\
  init=3: random configuration + soft sphere MC to remove overlaps\n\
          (fast, memory-consuming, not for low density)\n\
  init=4: old version of init=3 (deprecated)\n\
  init=5,6,7: new configuration sc,bcc,fcc\n\
  init>7: new configuration repeats rescaled unit cell stored as file of\n\
          x,y,z (0<x,y,z<1) with name \"SIMNAME.#\" (#=init)\n\
verbose: redirects output protocol to stdout instead of SIMNAME.prt\n\
  and overrides variable `verbose' in SIMNAME.get\n\
negative verbose: ignore the lock-file SIMNAME.loc\n",
            NSUBCELL,EPS,
#ifdef DEBUG
            DEBUG
#else
            (double)0
#endif
            );
#ifdef SWPERT
    fprintf(stderr,"SWPERT=%d version\n",SWPERT);
#endif            
    exit(0); }

  Fnarg1=arg[1];
  lock=fopen(Fn("loc"),"r+");
  if (narg>3) verbose=atoi(arg[3]);
  if (lock && verbose>=0)
    ERROR(("Lock file \"%s.loc\" exists\n\
(\"hsmd %s\" may be running or has not been stopped properly)",
          arg[1],arg[1]))
  else
    lock=fopen(Fn("loc"),"w");
  atexit(removelock);

#if CHECKHEAP==2
  if (narg>4) AllocRange=atoi(arg[4]);
  if (narg>5) AllocTrace=atoi(arg[5]);
  prt("CHECKHEAP %d %d",AllocRange,AllocTrace);
#endif
  if (narg<=3) {
    rename(fn=Fn("prt"),Fn("prt~"));
    out=fopen(fn,"wt");
    if (!out) ERROR(("cannot write to %s.prt",arg[1])) }

#ifdef __EMX__
  fprintf(stderr,"hsmd started, output protocol=%s\n",fn);
#else
  fprintf(stderr,"hsmd started, output protocol=%s\n\
interrupt (let finish cycle): ^C or `kill -INT', save all ASAP: `kill'\n",fn);
#endif

  in=fopen(fn=Fn("get"),"rt");
  if (!in) ERROR(("open %s",fn))

  /* info on NSUBCELL ... LONG also printed when help */
  prt("Compile-time options: NSUBCELL=%d EPS=%g DEBUG=%g"
#ifdef LONG
" LONG (rdf for r>box/2)"
#else
" !LONG (rdf for r<box/2-cell)"
#endif
      ,NSUBCELL,EPS,
#ifdef DEBUG
      DEBUG
#else
      (double)0
#endif
      );

  getdata
    get(rhos) get(rhof) /* const densities */
    get(x) get(n) get(V)
    get(cutoff)
    get(N)
/*.....   get(eta)*/
    get(rho)
    get(init) get(Tprotocol)
    getvec(LN,,DIM) get(rnucleus)
    get(Nlist) get(Nlistshift)
    get(E) get(slice)
    get(seed)
    get(ncell) get(nhash)
    get(rdf.grid) get(rdf.cutoff)
    get(dt.measure) get(dt.plb) get(dt.reset) get(dt.block) get(dt.cfg)
    get(dt.save) get(dt.stop)
    get(Zmin) get(Zmax)
    get(verbose)
    get(lag)
    get(AllocSizeLim)
#ifdef SWPERT
#if SWPERT<3
#define SWPERT 3
#endif
    get(rdf.nsw)
    if (rdf.nsw) {
      if (!rdf.swindex) {
        allocarrayzero(rdf.swindex,rdf.nsw+1);
        allocarrayzero(rdf.swsum,rdf.nsw);
        allocarrayzero(rdf.swpow,rdf.nsw); }
      getvec(rdf.swindex,,rdf.nsw+1) }
#endif
    checkdata enddata

  put3(N,rho,AllocSizeLim)

  it1=(int)(dt.block/dt.measure+0.5);
  if (fabs(dt.measure*it1-dt.block)>EPS*dt.measure)
    ERROR(("dt.block=%d must be an integer multiple of dt.measure=%d",
          dt.block,dt.measure))
  it1=(int)(dt.save/dt.measure+0.5);
  if (fabs(dt.measure*it1-dt.save)>EPS*dt.measure)
    WARNING(("for easier restart of crashed simulation,\n\
*** dt.save=%d should be an integer multiple of dt.measure=%d",
            dt.save,dt.measure))

  if (narg>2) init=atoi(arg[2]);
  if (narg>3) verbose=abs(atoi(arg[3]));

  if (cutoff<=1) ERROR(("wrong cutoff"))
  rndinit(7,seed);
  if (Nlist==0) Nlist=N;
  if (Nlistshift==0) Nlistshift=Nlist/2+1;
  put2(Nlist,Nlistshift)
  if (Nlist<4) ERROR(("minimum Nlist is 3")) /* see slice check! */
  if (Nlistshift<2) ERROR(("minimum Nlistshift is 1"))
  it1=Nlist;

  if (Zmax!=0 && Zmin!=0 && Zmin>=Zmax)
    ERROR(("[Zmin,Zmax] = [%g,%g] is not interval",Zmin,Zmax))

  if (dt.measure==0) {
    dt.measure=N;
    prt("dt.measure=N (approx. slice=%d collisions/sphere/measurement)",
        slice); }

/*..... rho=2*DIM/PI*eta;*/
  eta=PI/(2*DIM)*rho;
#if DIM==2
  L=sqrt(N/rho);
  Z=(1+2*eta)/(1-eta); /* wrong */
#else
  Z=(1+eta+eta*eta*(1-eta/1.5*(1+eta)))/Cub(1-eta);
  L=cubrt(N/rho);
#endif

  userL=L;

  prt("box size L=%.7f sigma, L/2=%.7f sigma, cutoff=%.7f sigma",L,L/2,cutoff);

  if (nhash==0) nhash=2*N;
  put(nhash)
  if (nhash<N/4 || nhash>10*N) WARNING(("nhash out of recommended range"))

  if (ncell==0) ncell=L*1.35;
  sigma=ncell/L; sigmaq=Sqr(sigma);
  put3(ncell,sigma,1/sigma)
  prt("av. cell occupancy=%.4f",(real)N/Cub(ncell));

#ifdef LONG
  rdfcq=Sqr(rdf.cutoff*sigma)*(1-EPS);
  /* to be modified later */
#else
  Lh=L/2*(1.-(1.+1./NSUBCELL)/ncell);
  if (rdf.cutoff>=Lh) {
    WARNING(("hsmd bug: rdf.cutoff too long, set to the algorithm limit %g",Lh))
    rdf.cutoff=Lh; }
  rdfcq=Sqr(rdf.cutoff*sigma);
#endif

  /* set E: slice ~ expected number of collisions per (first) time slice */
  if (E==0) {
    E=(PI/3)*Sqr(sigma*slice/(Z-1))/N;
    prt("E=%.12g set\n\
(E=2*Ekin gives approx slice=%d collisions/time slice)",E,slice); }
  L=ncell; Lh=ncell*0.5;
  cutoff*=sigma;

  makenbrs(nbrs,cutoff,0); /* may change cutoff */
#ifndef LONG
  makenbrs(rdfnbrs,rdf.cutoff*sigma,1);
#endif
  cq=Sqr(cutoff);
  collisionrange=(cutoff-sigma)/2;

  allocarrayzero(UPDATE,NUPDATE); /* to allow save events relocatably */
  MEASURE=UPDATE+1;
  PLB=UPDATE+2;
  CFG=UPDATE+3;
  RESET=UPDATE+4;
  BLOCK=UPDATE+5;
  SAVE=UPDATE+6;
  STOP=UPDATE+7;

  allocarray(cfg,N);

  allocarray(box,ncell);
  loop (i,0,ncell) {
    allocarray(box[i],ncell);
    loop (j,0,ncell) allocarrayzero(box[i][j],ncell); }

  allocarrayzero(tlist,Nlist+Nlistshift-1);
  makehash();

  rdf.nhist=rdf.grid*(rdf.cutoff+EPS/4);
  if (rdf.cutoff-(real)rdf.nhist/rdf.grid>EPS)
    WARNING(("rdf.cutoff rounded to %g to be an integer multiple of grid",
            rdf.cutoff=(real)rdf.nhist/rdf.grid))
  if (rdf.cutoff<1.1-EPS)
    ERROR(("rdf.cutoff must be at least 1.1 because of g1 convergence profile"))
  rdf.proggrid=rdf.grid/sigma;
#ifdef LONG
  /* adjusting rdfcq: no test for index */
  {
    int i;
    double dr=EPS*rdfcq,q;

    i=(int)(sqrt(rdfcq)*rdf.proggrid);
    if (i>=rdf.nhist) ERROR(("LONG: rdf.grid %d %d",i,rdf.nhist))
    loop (i,0,100) {
      q=rdfcq+dr;
      if ((int)(sqrt(rdfcq)*rdf.grid)<rdf.nhist) rdfcq=q;
      dr/=2; }
  }
#endif
  put2(rdf.nhist,rdf.proggrid)

#ifdef SWPERT
  if (!rdf.swindex)
    ERROR(("rdf.nsw not defined. (To run plain HS, #undef SWPERT)"))
  if (rdf.swindex[0]==0) rdf.swindex[0]=rdf.grid;
  if (rdf.swindex[0]!=rdf.grid) WARNING(("rdf.swindex[0]!=rdf.grid"))
  rdf.lag=lag;
  if (rdf.swindex[0]<0 || rdf.swindex[0]>rdf.nhist)
    ERROR(("rdf.swindex[0] out of range"))
  loopto (i,1,rdf.nsw) {
    prt("sw interval %d = [%g,%g)",i,
        rdf.swindex[0]/(real)rdf.grid,
        rdf.swindex[i]/(real)rdf.grid);
    if (rdf.swindex[i-1]>=rdf.swindex[i]) ERROR(("empty interval %d",i));
    if (rdf.swindex[i]>rdf.nhist) ERROR(("interval %d out of range",i)); }
#endif

  allocarrayzero(rdf.blockhist,rdf.nhist);
  allocarrayzero(rdf.tothist,rdf.nhist);

  put2(N,rnucleus)
  putv(LN)

  if (abs(init-1)<=1) loadall(Fn("all"),init);
  else initcfg(init,LN,rnucleus);

  if (init==0) StaLoad(Fn("sta"));
  else rename(Fn("res"),Fn("res~"));

  if (init>=2 || init<0) {
    if (t) ERROR(("t=%g (0 expected)",t))
    controlevent(dt.measure,MEASURE);
    controlevent(dt.plb,PLB);
    controlevent(dt.cfg,CFG);
    controlevent(dt.reset,RESET);
    controlevent(dt.block*(1+EPS),BLOCK);
    controlevent(dt.save*(1+EPS),SAVE); }

  if (dt.stop) {
    double dtstop=dt.stop*(1+EPS)-t-toffset;

    prt("stop in %g",dtstop);
    if (dtstop<=0)
      ERROR(("STOP expired (dtstop<=0)\n\
*** dt.stop=%d t=%g toffset=%g\n\
*** increase dt.stop (dt.stop=0 run forever)",dt.stop,t,toffset))

    controlevent(dtstop,STOP); }
  else
    prt("no stop");

  zeromomentum();
  if (fabs(setenergy(E)-1)>EPS)
    if (init==0 || init==1)
      ERROR(("sorry, can't change time units (E,slice,ncell) for init=0,1"))

  timeunit=sigma*sqrt(E/N);
  prt("\nConversion to comparable time scale units (TU):");
  prt("time slice width = %.12g Time Units",timeunit);
  prt("dt.measure = %.12g Time Units",dt.measure*timeunit);
  prt("(multiply line count [col.0] of %s.cp by this value)", arg[1]);
  if (dt.block) prt("dt.block = %.12g Time Units",dt.block*timeunit);
  if (dt.save) prt("dt.save = %.12g Time Units",dt.save*timeunit);
  if (dt.stop) prt("dt.stop = %.12g Time Units\n",dt.stop*timeunit);

  buildcells();
  makelists();

  if (dt.cfg)
    opencfg(Fn("cfg"),init);

  if (dt.plb) {
    openplb(Fn("plb"),init);
    if (init>=3 || init<0) {
      writeplb();
      prt("writing initial cfg, E=%.16g",energy()); } }

  opencp(Fn("cp"),init);

  signal(SIGINT,asksig);
  signal(SIGTERM,asksig);

#ifdef DEBUG
  ovltest("init");
#endif

  if (verbose) {
    liststat("hashed pair table",(void**)pair,nhash,NULL);
    liststat("event list",(void**)(tlist+it0),it1-it0,(void**)longt);
    cellstat();

    prt("\n\
-------- MD time ------                  Z (JK=%7.4f)        CPU time\n\
  [prog.units]    [TU]     err_E  colls  virial    rate  updt  MD  RDF",Z);
    clock0=clock(); }

  fflush(out);

  for ( ; sig!=SIGTERM; ) {
    struct event_s *e=nextevent();

    t=e->t;
    if (e->A==UPDATE) {
      collisions.update++;
      advance(e->B);
      collide(e->B,NULL);
      /* e is rescheduled within collide() */ }

    else if (e->A==PLB) {
      writeplb();
      rescheduleevent(dt.plb,e); }

    else if (e->A==CFG) {
      writecfg();
      rescheduleevent(dt.cfg,e); }

    else if (e->A==RESET) {
      real q;

      resettime();
      q=setenergy(E)-1;
      if (fabs(q)>20*EPS) ERROR(("energy error = %g",q))
      else if (fabs(q)>EPS) WARNING(("energy error = %g > %g",q,EPS))
      rescheduleevent(dt.reset,e); }

    else if (e->A==BLOCK) {
      appendblockg(Fn("res"),collisions.sumblock);
      collisions.sumblock=0;
      rescheduleevent(dt.block,e); }

    else if (e->A==STOP) {
      removeevent(e);
      break; }

    else if (e->A==SAVE) {
      liststat("hashed pair table",(void**)pair,nhash,NULL);
      liststat("event list",(void**)(tlist+it0),it1-it0,(void**)longt);
      cellstat();

      rescheduleevent(dt.save,e);
      saveall(Fn("alx"));
      StaSave(Fn("stx"));
      if (cp) fflush(cp);
      prt_("SAVED: %s.alx %s.stx\nFLUSHED:",arg[1],arg[1]);
      if (plb) prt_(" %s.plb (%ld)",arg[1],ftell(plb));
      if (cp) prt_(" %s.cp (%ld)",arg[1],ftell(cp));
      _n
      prt("==========================================================================");
      fflush(out); }

    else if (e->A==MEASURE) {
      real Z_vir = 1 - (N-1)/(double)N/(E*dt.measure)*collisions.virial;
      real Z_rate = 1 + gausscorr(N)
                        *sigma/sqrt(3/PI*N*E)/dt.measure*collisions.block;
      real locE=energy();
      real Eerr=locE/E-1;
      real updateratio=collisions.update/(collisions.block+1e-9);
      real g1;
      clock_t clockrdf=0;
      FILE *stp=fopen(Fn("stp"),"r");

      if (stp) {
        sig=SIGINT;
        prt("%s detected ==> SIGINT signalled",Fn("stp"));
        fclose(stp);
        remove(Fn("stp")); }

      if (verbose) clockrdf=clock();
      g1=measurerdf();

      writecp((t+toffset)*timeunit,Z_vir,Z_rate,g1,updateratio);

      if (collisions.update && collisions.update>9)
        updatewarning(updateratio);
      slicewarning();

      StaSet(lag,2,8);
      StaAdd("z vir",Z_vir);
      StaAdd("z rate",Z_rate);
      StaSet(2,0,0);
      StaAdd("Eerr",Eerr);
      StaSet(2,2,8);
      loopto (i,1,13) if (i!=10)
        StaAdd(string("z 0.%dvir+0.%drate",10-i,i),(Z_vir*(10-i)+Z_rate*i)/10);
      StaAdd("g1",g1);

      if (Zmin!=0) {
        if (Z_rate<Zmin && Z_vir<Zmin) {
          if (!sig) prt("Z > Zmin ==> SIGINT signalled",Zmin);
          sig=SIGINT; } }
      if (Zmax!=0) {
        if (Z_rate>Zmax && Z_vir>Zmax) {
          if (!sig) prt("Z > Zmax ==> SIGINT signalled",Zmax);
          sig=SIGINT; } }

      if (verbose) {
        static int nc=0;
        clockt=clock();
        lastcpu=(double)(clockrdf-clock0)/CLOCKS_PER_SEC;
        prt("%11.6g %11.6g %9.2g %5d %7.4f %7.4f %5.3f %.2f %.2f",
            t,(toffset+t)*timeunit,Eerr,collisions.block,
            Z_rate,Z_vir,
            updateratio,
            lastcpu,
            (double)(clockt-clockrdf)/CLOCKS_PER_SEC);
        if (nc) sumcpubutfirst+=lastcpu;
        nc++;
        clock0=clockt; }

      collisions.sumblock+=collisions.block;
      {
        double cpera=collisions.block/(double)N;
        static int pass;

        pass++;
        if (pass>4 && pass<30 && (cpera<1 || cpera>25)) {
          pass=30;
          WARNING(("Inefficient # of collisions/atom. Check dt.measure and slice")) }
        StaSet(2,0,0);
        StaAdd("colls/atom/blk",collisions.block/(double)N);
      }
      collisions.block=0;
      collisions.update=0;
      collisions.virial=0;
      rescheduleevent(dt.measure,e);
#ifdef __EMX__
      if (!verbose) fprintf(stderr,"\r%g\t(to interrupt: `!'=ASAP, ESC=finish cycle)\t",t);
      if (_read_kbd(0,0,0)==27) {
        sig++;
        fprintf(stderr,"\nESC hit, wait...\n"); }
#endif
      if (sig) break;
      }

    else { /* COLLISION */
      struct HS_s *A=e->A,*B=e->B;
/*.....prt("~ %g %d %d",e->t,e->A-cfg,e->B-cfg);*/

      if (t!=A->t) advance(A);
      if (t!=B->t) advance(B);
      collision(A,B);
      updatecollisionlist(NOTIME,A,B);
      collide(A,B);
      collide(B,A);
      collisions.block++;
      /* e is rescheduled within collide() */ }
    }

  if (cfgf) fclose(cfgf);
  if (plb) fclose(plb);
  if (cp) closecp();

  if (verbose) {
    prt("SUMMARY ncell cutoff slice CPU:");
    if (sumcpubutfirst)
      prt("%d %g %g %g CPU TIME BUT 1st CYCLE",
          ncell,cutoff/sigma,slice,sumcpubutfirst);
    else
      prt("%d %g %g %g CPU TIME 1st CYCLE",
          ncell,cutoff/sigma,slice,lastcpu); }

  /* backups */
  {
    char *to;

    fn=Fn("all");
    to=Fn("all~");
    if (!rename(fn,to)) if (verbose) prt("%s renamed to %s",fn,to);
    saveall(fn);

    fn=Fn("sta");
    to=Fn("sta~");
    if (!rename(fn,to)) if (verbose) prt("%s renamed to %s",fn,to);
    StaSave(fn);
  }

#ifdef DEBUG
  checklist("final",0);
  ovltest("finish");
#endif

  StaPrintAll(NULL);

#ifdef SWPERT
  {
    FILE *f=fopen(Fn("res"),"rt");
    char line[1024];
    int i,ii,iw,ip,n;
    double (*S)[SWPERT],(*Q)[SWPERT];
    int nblk=0;
    double x;
    FILE *pert[SWPERT];

    prt_("lambda0=%g\n  lambda\t",rdf.swindex[0]/(real)rdf.grid);
    loopto (ip,1,SWPERT) {
      prt_("<(sw bonds/atom)^%d>\t",ip);
      pert[ip-1]=fopen(Fn(string("pert%d",ip)),"wt"); }
    _n
    loopto (i,1,rdf.nsw) {
      prt_("%9g", rdf.swindex[i]/(real)rdf.grid);
      loopto (ip,1,SWPERT) prt_("\t%17.15g",StaMean(string("sw[%d]^%d",i,ip)));
      _n }

    allocarray(S,rdf.nsw);
    allocarray(Q,rdf.nsw);
    loop (i,0,SWPERT) loop (ii,0,rdf.nsw) S[ii][i]=Q[ii][i]=0;

    if (!f) {
      WARNING(("no %s - cannot calculate errors of perturbation terms",Fn("res")))
      goto ex; }

    for (n=0;; n++) {
      int br=0;

      do if (!fgets(line,1024,f)) { br=1; break; }
      while (line[0]!='#');
      if (br) break;

      i=atoi(line+2);
      if (i!=N) {
        WARNING(("wrong N in %s",Fn("res")))
        goto ex; }
      while (line[0]=='#') {
        if (!fgets(line,1024,f)) {
          WARNING(("wrong %s",Fn("res")))
          goto ex; } }

      nblk++;

      loop (iw,0,rdf.nsw) {
        char *tok=strtok(line," \t\n");
        tok=strtok(NULL," \t\n");
        tok=strtok(NULL," \t\n");
        tok=strtok(NULL," \t\n");
        loop (ip,0,SWPERT) {
          tok=strtok(NULL," \t\n");
          if (!tok) {
            WARNING(("wrong format of %s",Fn("res")))
            goto ex; }
          x=atof(tok);
          S[iw][ip]+=x;
          Q[iw][ip]+=x*x; }
        if (!fgets(line,1024,f)) {
          WARNING(("wrong %s",Fn("res")))
          goto ex; }
        } }

    ex:
    _n put(nblk)
    if (f) {
      fclose(f);
      if (nblk<2) WARNING(("not enough blocks for error calculation")) }

    /* will print warnings for SWPERT<6 */
/***
  perturbation terms by derive, a=<Delta U>, b=<Delta U^2>, etc; t=1/(kT)
  generator function: - ln <exp(-t Delta U)>
  t: a
  t^2: a^2/2-b/2
  t^3: a^3/3-a*b/2+c/6
  t^4: a^4/4-a^2*b/2+a*c/6+b^2/8-d/24
  t^5: a^5/5-a^3*b/2+a^2*c/6+a*b^2/4-a*d/24-b*c/12+e/120
  t^6: a^6/6-a^4*b/2+a^3*c/6+3*a^2*b^2/8-a^2*d/24-a*b*c/6+a*e/120-b^3/24+b*d/48+c^2/72-f/720
***/

    prt_("\nWARNING: values in the \"by blocks\" column may be considerably biased!\n  lambda\t");
    loopto (ip,1,SWPERT) prt_("\t--------- Apert%d/beta^%d ----------\t",ip,ip);
    prt_("\n        \t");
    loopto (ip,1,SWPERT) prt_(" whole-run average\tby blocks\tstderr\t");
    _n

  loop (ip,0,SWPERT)
    fprintf(pert[ip],"# N=%d rho=%g nblk=%d\n# WARNING: values in the \"by blocks\" column may be considerably biased!\n#  lambda\t\t --------- Apert%d/beta^%d ---------- \n#        \t whole-run average   by blocks  \tstderr\n",N,rho,nblk,ip+1,ip+1);

  loopto (i,1,rdf.nsw) {
      real a=StaMean(string("sw[%d]^%d",i,1));
      real b=StaMean(string("sw[%d]^%d",i,2));
      real c=StaMean(string("sw[%d]^%d",i,3));
#if SWPERT>3
      real d=StaMean(string("sw[%d]^%d",i,4));
#if SWPERT>4
      real e=StaMean(string("sw[%d]^%d",i,5));
#if SWPERT>5
      real f=StaMean(string("sw[%d]^%d",i,6));
#if SWPERT>6
#error SWPERT>6 not supported
#endif
#endif
#endif
#endif

      prt_("%9g", rdf.swindex[i]/(real)rdf.grid);
      loop (ip,0,SWPERT) fprintf(pert[ip],"%9g", rdf.swindex[i]/(real)rdf.grid);

#define AVERR(I) (S[i-1][I]/nblk),(sqrt((Q[i-1][I]/nblk-Sqr(S[i-1][I]/nblk))/(nblk-1)))

      prt_("\t%15.8f\t%15.8f\t%15.8f",a,AVERR(0));
      fprintf(pert[0],"\t%15.8f\t%15.8f\t%15.8f\n",a,AVERR(0));

      prt_("\t%15.8f\t%15.8f\t%15.8f",(powi(a,2)/2-b/2)*N,AVERR(1));
      fprintf(pert[1],"\t%15.8f\t%15.8f\t%15.8f\n",(powi(a,2)/2-b/2)*N,AVERR(1));

      prt_("\t%15.8f\t%15.8f\t%15.8f",(powi(a,3)/3-a*b/2+c/6)*N*N,AVERR(2));
      fprintf(pert[2],"\t%15.8f\t%15.8f\t%15.8f\n",(powi(a,3)/3-a*b/2+c/6)*N*N,AVERR(2));
#if SWPERT>3
      prt_("\t%15.8f\t%15.8f\t%15.8f",(powi(a,4)/4-powi(a,2)*b/2+a*c/6+powi(b,2)/8-d/24)*N*N*N,AVERR(3));
      fprintf(pert[3],"\t%15.8f\t%15.8f\t%15.8f\n",(powi(a,4)/4-powi(a,2)*b/2+a*c/6+powi(b,2)/8-d/24)*N*N*N,AVERR(3));
#if SWPERT>4
      prt_("\t%15.8g\t%15.8g\t%15.8g",(powi(a,5)/5-powi(a,3)*b/2+powi(a,2)*c/6+a*powi(b,2)/4-a*d/24-b*c/12+e/120)*N*N*N*N,AVERR(4));
      fprintf(pert[4],"\t%15.8g\t%15.8g\t%15.8g\n",(powi(a,5)/5-powi(a,3)*b/2+powi(a,2)*c/6+a*powi(b,2)/4-a*d/24-b*c/12+e/120)*N*N*N*N,AVERR(4));
#if SWPERT>5
      prt_("\t%15.8g\t%15.8g\t%15.8g",(powi(a,6)/6-powi(a,4)*b/2+powi(a,3)*c/6+3*powi(a,2)*powi(b,2)/8-powi(a,2)*d/24-a*b*c/6+a*e/120-powi(b,3)/24+b*d/48+powi(c,2)/72-f/720)*N*N*N*N*N,AVERR(5));
      fprintf(pert[5],"\t%15.8g\t%15.8g\t%15.8g\n",(powi(a,6)/6-powi(a,4)*b/2+powi(a,3)*c/6+3*powi(a,2)*powi(b,2)/8-powi(a,2)*d/24-a*b*c/6+a*e/120-powi(b,3)/24+b*d/48+powi(c,2)/72-f/720)*N*N*N*N*N,AVERR(5));
#endif
#endif
#endif
      _n }
    loop (ip,0,SWPERT) fclose(pert[ip]);
    _n
  }
#endif /* SWPERT */

  prt("Z[JKEOS]=%g  <l>/l0 = %g",Z,4*eta/StaMean("z rate"));

  writetotg(Fn("g"));

  liststat("hashed pair table",(void**)pair,nhash,NULL);
  liststat("event list",(void**)(tlist+it0),it1-it0,(void**)longt);
  cellstat();

  prt("STOP: t=%g  toffset=%g  total=%g TU",t,toffset,(toffset+t)*timeunit);

  return 0;
}
