#include <limits.h>

struct event_s {
  struct event_s *next,*prev;
  real t; /* event time (from toffset) */
  struct HS_s *A,*B; /* pointers to 2 spheres that are going to collide;
                        A=UPDATE: this is the `update' event for sphere B
                        A=MEASURE,PLB,RESET,BLOCK:
                            other events  (B irrelevant) */
  }
  **tlist, /* tlist[it] = head of event list containing times [it,it+1) */
  *longt;  /* head of events scheduled for longer times 
              than the last element of tlist */

int it0,it1; /* sliding bounds of tlist (NOTE: tlist slides, too) */

int nhash;

struct pair_s {
  struct pair_s *next;
  struct HS_s *A,*B; /* the two spheres */ 
  struct event_s *l; /* pointer to the tlist table */
  real t; /* time of event (Q:IS NEEDED?) */ 
  } **pair; /* array[nhash]: index is the sum of hash values */

void removeevent(struct event_s *e) /************************** removeevent */
/* 
  removes event e from the list but does not erase it 
  (i.e., does not free the memory so that the event can be re-used)
*/
{
#ifdef DEBUG
  if (!e) ERROR(("no event to remove"))
#endif
  if (e->prev) {
    e->prev->next=e->next;
    if (e->next) e->next->prev=e->prev; }
  else {
    /* no previous - it must be the head */
    int it=(int)e->t;
    struct event_s **remhead = it>=it1 ? &longt : &tlist[it];

#ifdef DEBUG
    if (e->t==NOTIME) ERROR(("NOTIME in the list"))
#endif

    if (*remhead!=e) {
     /* this should never happen ... but rounding while it=e->t */
      int i,found=-1;
      remhead=NULL;

      if (e==longt) {
        found=it1;
        remhead=&longt; }
      else
        loop (i,it0,it1) if (e==tlist[i]) {
          if (remhead)
            ERROR(("t=%.13g e->t=%.13g event twice in tlist",t,e->t))
          remhead=&tlist[found=i]; }
      if (!remhead)
        ERROR(("t=%.13g e->t=%.13g event not found",t,e->t))
       
      WARNING(("event misplaced in tlist\n\
*** t=%.16g e->t=%.16g: it=%d but found at it=%d", t,e->t,it,found))

      if (abs(it-found)>1) ERROR(("displaced too much")) }

    *remhead=(*remhead)->next;
    if (*remhead) (*remhead)->prev=NULL; }
}

void scheduleevent(real tt,struct event_s *e) /*************** scheduleevent */
/* 
  schedule event e at time t+tt
  does not remove the previous instance (use removeevent)
  don't use for collision events (use updatecollisionlist() instead)
*/
{
  int it;
  struct event_s **head;

#ifdef DEBUG
  if (tt==NOTIME) ERROR(("t=NOTIME"))
  if (tt<0) ERROR(("tt=%g<0",tt))
#endif

  it = tt += t;

  if (tt>(double)INT_MAX)
    ERROR(("event scheduler implementation limit (%g) exceeded\n\
*** (time of scheduled event = %g)",(double)INT_MAX,tt))

  if (it<it1) head=&tlist[it];
  else head=&longt;

  e->t=tt;

  /* including event e into tlist */
  e->next=*head;
  e->prev=NULL;
  if (*head) (*head)->prev=e;
  *head=e;
}

void rescheduleevent(real tt,struct event_s *e) /********** rescheduleevent */
{
  removeevent(e);
  scheduleevent(tt,e);
}

void cellstat(void) /********************************************* cellstat */
{
  intvector ir;
  struct HS_s *A;
#define NH 8
  int hist[NH],more=0;
  int h,NHS=0;

  loop (h,0,NH) hist[h]=0;

  loop (ir[0],0,ncell)
    loop (ir[1],0,ncell)
      loop (ir[2],0,ncell) {
        h=0;
        looplist (A,box[ir[0]][ir[1]][ir[2]]) { h++; NHS++; }
        if (h<NH) hist[h]++; else more++; }

  prt_("box list:");
  loop (h,0,NH) prt_(" %d*%d +",h,hist[h]);
  prt(" more*%d",more);
  if (Ntot!=NHS) ERROR(("HS lost: NHS=%d != N=%d",NHS,Ntot));
#undef NH
}

void advance(struct HS_s *A) /************************************* advance */
/*
  advances A->r[] so that its local time A->t matches the global time t
  no collisions assumed in between!
  cell updated if necessary
  not supposed to be called for t=A->t (not wrong, just inefficient)
*/
{
  double tt=t-A->t;
  int k,cellchanged;
  intvector ir;

#ifdef DEBUG
//prt("%2d: t=%.20f A=%.20f %.20f", A-cfg,t,A->t,tt);

  if (tt<0) ERROR(("t=%g < t->(%d)=%g (dif=%g)",t,A-cfg,A->t,tt))

  {
    int k;

    if (A->cell!=&box[A->ir[0]][A->ir[1]][A->ir[2]]) ERROR(("A->cell wrong"))

    if (!box[A->ir[0]][A->ir[1]][A->ir[2]]) {
      WARNING(("NULL in a box that should contain a sphere"))
      prtsphere(A); }

    loop (k,0,DIM) 
      if (A->r[k] < A->ir[k]-DEBUG || A->ir[k] > A->ir[k]+(1+DEBUG))
          ERROR(("A->ir=%d %d %d does not match A->r=%g %g %g",
                 A->ir[0],A->ir[1],A->ir[2],
                 A->r[0], A->r[1], A->r[2]))
    }
#endif

  A->t=t;
  VV(A->r,+=tt*A->v)

  cellchanged=0;
  loop (k,0,DIM) {
    ir[k]=(int)(A->r[k]+L)%ncell;
    if (ir[k]!=A->ir[k]) {
      cellchanged++;
      if (A->r[k]>=L) A->r[k]-=L;
      if (A->r[k]<0) A->r[k]+=L; 
      A->ir[k]=(int)A->r[k];
      /* ... of course, it should be (int)A->r[k] == ir[k]
         but, because of rounding, it might rarely differ, which would cause 
         problems in collide() and measurerdf() */
      } }

  if (cellchanged) {
    struct HS_s **newcell=&box[A->ir[0]][A->ir[1]][A->ir[2]],**XX,*X;

    /* removing HS from the linked list */
    for (XX=A->cell,X=*XX; X; XX=&X->next,X=*XX)
      if (X==A) { *XX=X->next; goto removed; }
    ERROR(("sphere %d %d %d not in its list",A->ir[0],A->ir[1],A->ir[2]))
   removed:

    /* adding HS to the new list */
    A->cell=newcell; 
    A->next=*newcell;
    *newcell=A; }
}

struct event_s *nextevent(void) /******************************** nextevent */
/*
  determines the next event from the time-ordered list of events
  rearranges the list of events if necessary
*/
{
  int ic=0;
  struct event_s *l,*lmintime=NULL;
  static int tlistshift=0;

  while (!tlist[it0]) {
    int it;
    struct event_s *newlongt=NULL,*newend=NULL,*lnext,**l0;
    real tmax;

/*.....prt("it0=%d it1=%d will be advanced",it0,it1);*/

    if (ic++>=Nlist)
      ERROR(("empty collision time list, check Nlist, dtlist, cutoff"))
    it0++;
    if (++tlistshift>=Nlistshift) {
      /* moving the apparent array within a (longer) total array */
      tlist-=Nlistshift; /* dirty: array offset outside array */
      loop (it,it0,it1) tlist[it]=tlist[it+Nlistshift]; 
      tlistshift=0; }
    it1++;

    tmax=it1;
    for (l=longt; l; l=lnext) {
      lnext=l->next;
      if (l->t>=tmax) l0=&newlongt; else l0=&newend;
      l->next=*l0;
      l->prev=NULL;
      if (*l0) (*l0)->prev=l;
      *l0=l; }

    tlist[it1-1]=newend;
    longt=newlongt; 

#ifdef DEBUG
    checklist("advlist",0);
#endif
    }

  /* leaving this test even with no DEBUG */
  if (!tlist[it0]) {
    ERROR(("nextevent: no event"))
    return NULL; }

  lmintime=tlist[it0];
  looplist (l,lmintime->next)
    if (l->t<lmintime->t) lmintime=l;

  return lmintime;
}

static real lastsynctime=-9e9;

void syncHS(void) /************************************************* syncHS */
/*
  advances all HS so that their local time = global time t
*/
{
  int i;

  if (t==lastsynctime) return;

  loop (i,0,Ntot) if (cfg[i].t!=t) advance(&cfg[i]);
  lastsynctime=t;
}

void liststat(char *title,void **array,int len,void **longt) /**** liststat */
/* 
  statistics of lists; length of 1st element returned 
*/
{
  int i,null=0,maxl=0,sum=0,n=-1;
  struct list_s {
    struct list_s *next; } *l;

  prt_("%-17s:",title);

  loop (i,0,len) {
    n=0;
    looplist (l,array[i]) n++;
    if (!n) null++;
    Max(maxl,n)
    sum+=n; 
    if (i<16) prt_(" %d",n); }
  prt_(" ... %d",n);

  if (longt) {
    n=0;
    looplist (l,*longt) n++;
    prt_(" long=%d",n); }

  prt("\n  array size=%d, %d NULLs, average list length=%.3f, max=%d",
      len,null,(double)sum/len,maxl);
}

